@@ -50,7 +50,7 @@
 #endif
 #if HAVE_GNUMALLOC_H
 #include <gnumalloc.h>
-#elif HAVE_MALLOC_H && !defined(_SQUID_FREEBSD_) && !defined(_SQUID_NEXT_)
+#elif HAVE_MALLOC_H
 #include <malloc.h>
 #endif
 #if HAVE_MEMORY_H
@@ -99,8 +99,6 @@
 /* #define DEBUG_VARS_DECODE 1 */
 /* #define DEBUG_VARS_ENCODE 1 */
 
-#define ASN_PARSE_ERROR(x) { return(x); }
-
 /* Create a new variable_list structure representing oid Name of length Len.
  *
  * Returns NULL upon error.
@@ -115,7 +113,8 @@ snmp_var_new(oid * Name, int Len)
     printf("VARS: Creating.\n");
 #endif
 
-    New = (struct variable_list *) xmalloc(sizeof(struct variable_list));
+    New = xmalloc(sizeof(*New));
+    /* XXX xmalloc never returns NULL */
     if (New == NULL) {
 	snmp_set_api_error(SNMPERR_OS_ERR);
 	return (NULL);
@@ -131,6 +130,7 @@ snmp_var_new(oid * Name, int Len)
 	return (New);
     }
     New->name = (oid *) xmalloc(Len * sizeof(oid));
+    /* XXX xmalloc never returns NULL */
     if (New->name == NULL) {
 	xfree(New);
 	snmp_set_api_error(SNMPERR_OS_ERR);
@@ -147,6 +147,17 @@ snmp_var_new(oid * Name, int Len)
     return (New);
 }
 
+struct variable_list *
+snmp_var_new_integer(oid * Name, int Len, int ival, unsigned char type)
+{
+    variable_list *v = snmp_var_new(Name, Len);
+    v->val_len = sizeof(int);
+    v->val.integer = xmalloc(sizeof(int));
+    v->type = type;
+    *(v->val.integer) = ival;
+    return v;
+}
+
 /* Clone a variable list.
  *
  * Returns NULL upon error.
@@ -214,15 +225,15 @@ snmp_var_clone(struct variable_list *Src)
 
 /* Free a variable_list.
  */
-void 
+void
 snmp_var_free(struct variable_list *Ptr)
 {
-    if (Ptr->name && Ptr->name_length > 0)
+    if (Ptr->name)
 	xfree((char *) Ptr->name);
 
-    if (Ptr->val.string && Ptr->val_len > 0)
+    if (Ptr->val.string)
 	xfree((char *) Ptr->val.string);
-    else if (Ptr->val.integer && Ptr->val_len > 0)
+    else if (Ptr->val.integer)
 	xfree((char *) Ptr->val.integer);
 
     xfree(Ptr);
@@ -363,7 +374,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
     struct variable_list ** VarP,
     int Version)
 {
-    struct variable_list *Var, **VarLastP;
+    struct variable_list *Var = NULL, **VarLastP;
     u_char *bufp, *tmp;
     u_char VarBindType;
     u_char *DataPtr;
@@ -381,16 +392,18 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
     /* Now parse the variables */
     bufp = asn_parse_header(Buffer, &AllVarLen, &VarBindType);
     if (bufp == NULL)
-	ASN_PARSE_ERROR(NULL);
+	return (NULL);
 
     if (VarBindType != (u_char) (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
 	snmp_set_api_error(SNMPERR_PDU_PARSE);
-	ASN_PARSE_ERROR(NULL);
+	return (NULL);
     }
 #ifdef DEBUG_VARS_DECODE
     printf("VARS: All Variable length %d\n", AllVarLen);
 #endif
 
+#define PARSE_ERROR { snmp_var_free(Var); return(NULL); }
+
     /* We know how long the variable list is.  Parse it. */
     while ((int) AllVarLen > 0) {
 
@@ -403,7 +416,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	ThisVarLen = AllVarLen;
 	tmp = asn_parse_header(bufp, &ThisVarLen, &VarBindType);
 	if (tmp == NULL)
-	    ASN_PARSE_ERROR(NULL);
+	    PARSE_ERROR;
 
 	/* Now that we know the length , figure out how it relates to 
 	 * the entire variable list
@@ -414,7 +427,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	/* Is it valid? */
 	if (VarBindType != (u_char) (ASN_SEQUENCE | ASN_CONSTRUCTOR)) {
 	    snmp_set_api_error(SNMPERR_PDU_PARSE);
-	    ASN_PARSE_ERROR(NULL);
+	    PARSE_ERROR;
 	}
 #ifdef DEBUG_VARS_DECODE
 	printf("VARS: Header type 0x%x (%d bytes left)\n", VarBindType, ThisVarLen);
@@ -424,13 +437,13 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	bufp = asn_parse_objid(bufp, &ThisVarLen, &VarBindType,
 	    Var->name, &(Var->name_length));
 	if (bufp == NULL)
-	    ASN_PARSE_ERROR(NULL);
+	    PARSE_ERROR;
 
 	if (VarBindType != (u_char) (ASN_UNIVERSAL |
 		ASN_PRIMITIVE |
 		ASN_OBJECT_ID)) {
 	    snmp_set_api_error(SNMPERR_PDU_PARSE);
-	    ASN_PARSE_ERROR(NULL);
+	    PARSE_ERROR;
 	}
 #ifdef DEBUG_VARS_DECODE
 	printf("VARS: Decoded OBJID (%d bytes). (%d bytes left)\n",
@@ -444,7 +457,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	/* find out type of object */
 	bufp = asn_parse_header(bufp, &ThisVarLen, &(Var->type));
 	if (bufp == NULL)
-	    ASN_PARSE_ERROR(NULL);
+	    PARSE_ERROR;
 	ThisVarLen = DataLen;
 
 #ifdef DEBUG_VARS_DECODE
@@ -459,7 +472,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	    Var->val.integer = (int *) xmalloc(sizeof(int));
 	    if (Var->val.integer == NULL) {
 		snmp_set_api_error(SNMPERR_OS_ERR);
-		return (NULL);
+		PARSE_ERROR;
 	    }
 	    Var->val_len = sizeof(int);
 	    bufp = asn_parse_int(DataPtr, &ThisVarLen,
@@ -478,7 +491,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	    Var->val.integer = (int *) xmalloc(sizeof(u_int));
 	    if (Var->val.integer == NULL) {
 		snmp_set_api_error(SNMPERR_OS_ERR);
-		return (NULL);
+		PARSE_ERROR;
 	    }
 	    Var->val_len = sizeof(u_int);
 	    bufp = asn_parse_unsigned_int(DataPtr, &ThisVarLen,
@@ -497,7 +510,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	    Var->val.string = (u_char *) xmalloc((unsigned) Var->val_len);
 	    if (Var->val.string == NULL) {
 		snmp_set_api_error(SNMPERR_OS_ERR);
-		return (NULL);
+		PARSE_ERROR;
 	    }
 	    bufp = asn_parse_string(DataPtr, &ThisVarLen,
 		&Var->type, Var->val.string,
@@ -516,7 +529,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	    Var->val.objid = (oid *) xmalloc((unsigned) Var->val_len);
 	    if (Var->val.integer == NULL) {
 		snmp_set_api_error(SNMPERR_OS_ERR);
-		return (NULL);
+		PARSE_ERROR;
 	    }
 	    /* Only copy if we successfully decoded something */
 	    if (bufp) {
@@ -537,18 +550,16 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	case SMI_COUNTER64:
 	    snmplib_debug(2, "Unable to parse type SMI_COUNTER64!\n");
 	    snmp_set_api_error(SNMPERR_UNSUPPORTED_TYPE);
-	    return (NULL);
-	    break;
+	    PARSE_ERROR;
 
 	default:
 	    snmplib_debug(2, "bad type returned (%x)\n", Var->type);
 	    snmp_set_api_error(SNMPERR_PDU_PARSE);
-	    return (NULL);
-	    break;
+	    PARSE_ERROR;
 	}			/* End of var type switch */
 
 	if (bufp == NULL)
-	    return (NULL);
+	    PARSE_ERROR;
 
 #ifdef DEBUG_VARS_DECODE
 	printf("VARS:  Adding to list.\n");
@@ -557,6 +568,7 @@ snmp_var_DecodeVarBind(u_char * Buffer, int *BufLen,
 	*VarLastP = Var;
 	VarLastP = &(Var->next_variable);
     }
+#undef PARSE_ERROR
 
     return (bufp);
 }
