@@ -1,21 +1,21 @@
 
 /*
- * $Id: HttpRequest.c,v 1.23 1999/01/29 23:39:11 wessels Exp $
+ * $Id: HttpRequest.c,v 1.40 2006/06/06 05:57:55 hno Exp $
  *
  * DEBUG: section 73    HTTP Request
  * AUTHOR: Duane Wessels
  *
- * SQUID Internet Object Cache  http://squid.nlanr.net/Squid/
+ * SQUID Web Proxy Cache          http://www.squid-cache.org/
  * ----------------------------------------------------------
  *
- *  Squid is the result of efforts by numerous individuals from the
- *  Internet community.  Development is led by Duane Wessels of the
- *  National Laboratory for Applied Network Research and funded by the
- *  National Science Foundation.  Squid is Copyrighted (C) 1998 by
- *  Duane Wessels and the University of California San Diego.  Please
- *  see the COPYRIGHT file for full details.  Squid incorporates
- *  software developed and/or copyrighted by other sources.  Please see
- *  the CREDITS file for full details.
+ *  Squid is the result of efforts by numerous individuals from
+ *  the Internet community; see the CONTRIBUTORS file for full
+ *  details.   Many organizations have provided support for Squid's
+ *  development; see the SPONSORS file for full details.  Squid is
+ *  Copyrighted (C) 2001 by the Regents of the University of
+ *  California; see the COPYRIGHT file for full details.  Squid
+ *  incorporates software developed and/or copyrighted by other
+ *  sources; see the CREDITS file for full details.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -44,6 +44,9 @@ requestCreate(method_t method, protocol_t protocol, const char *urlpath)
     if (urlpath)
 	stringReset(&req->urlpath, urlpath);
     req->max_forwards = -1;
+    req->lastmod = -1;
+    req->client_addr = no_addr;
+    req->my_addr = no_addr;
     httpHeaderInit(&req->header, hoRequest);
     return req;
 }
@@ -52,14 +55,34 @@ void
 requestDestroy(request_t * req)
 {
     assert(req);
-    safe_free(req->body);
+    if (req->body_reader)
+	requestAbortBody(req);
+    if (req->auth_user_request)
+	authenticateAuthUserRequestUnlock(req->auth_user_request);
     safe_free(req->canonical);
+    safe_free(req->vary_hdr);
+    safe_free(req->vary_headers);
+    safe_free(req->urlgroup);
+    safe_free(req->extacl_user);
+    safe_free(req->extacl_passwd);
     stringClean(&req->urlpath);
     httpHeaderClean(&req->header);
     if (req->cache_control)
 	httpHdrCcDestroy(req->cache_control);
     if (req->range)
 	httpHdrRangeDestroy(req->range);
+    stringClean(&req->extacl_log);
+    if (req->vary) {
+	if (req->etags == &req->vary->etags)
+	    req->etags = NULL;
+	storeLocateVaryDone(req->vary);
+	req->vary = NULL;
+    }
+    assert(req->etags == NULL);
+    safe_free(req->etag);
+    if (req->pinned_connection)
+	cbdataUnlock(req->pinned_connection);
+    req->pinned_connection = NULL;
     memFree(req, MEM_REQUEST_T);
 }
 
@@ -91,19 +114,8 @@ httpRequestParseHeader(request_t * req, const char *parse_start)
     return httpHeaderParse(&req->header, blk_start, blk_end);
 }
 
-/* swaps out request using httpRequestPack */
-void
-httpRequestSwapOut(const request_t * req, StoreEntry * e)
-{
-    Packer p;
-    assert(req && e);
-    packerToStoreInit(&p, e);
-    httpRequestPack(req, &p);
-    packerClean(&p);
-}
-
 /* packs request-line and headers, appends <crlf> terminator */
-void
+static void
 httpRequestPack(const request_t * req, Packer * p)
 {
     assert(req && p);
@@ -116,6 +128,37 @@ httpRequestPack(const request_t * req, Packer * p)
     packerAppend(p, "\r\n", 2);
 }
 
+/* packs debug info, canonical request-line and headers, appends <crlf> terminator */
+void
+httpRequestPackDebug(request_t * req, Packer * p)
+{
+    assert(req && p);
+    /* Client info */
+    packerPrintf(p, "Client: %s ", inet_ntoa(req->client_addr));
+    packerPrintf(p, "http_port: %s:%d", inet_ntoa(req->my_addr), req->my_port);
+    if (req->auth_user_request && authenticateUserRequestUsername(req->auth_user_request))
+	packerPrintf(p, "user: %s", authenticateUserRequestUsername(req->auth_user_request));
+    packerPrintf(p, "\n");
+    /* pack request-line */
+    packerPrintf(p, "%s %s HTTP/%d.%d\r\n",
+	RequestMethodStr[req->method], urlCanonical(req), req->http_ver.major, req->http_ver.minor);
+    /* headers */
+    httpHeaderPackInto(&req->header, p);
+    /* trailer */
+    packerAppend(p, "\r\n", 2);
+}
+
+/* swaps out request using httpRequestPack */
+void
+httpRequestSwapOut(const request_t * req, StoreEntry * e)
+{
+    Packer p;
+    assert(req && e);
+    packerToStoreInit(&p, e);
+    httpRequestPack(req, &p);
+    packerClean(&p);
+}
+
 #if UNUSED_CODE
 void
 httpRequestSetHeaders(request_t * req, method_t method, const char *uri, const char *header_str)
@@ -138,16 +181,54 @@ httpRequestPrefixLen(const request_t * req)
 	req->header.len + 2;
 }
 
-/* returns true if header is allowed to be passed on */
+/*
+ * Returns true if HTTP allows us to pass this header on.  Does not
+ * check anonymizer (aka header_access) configuration.
+ */
 int
 httpRequestHdrAllowed(const HttpHeaderEntry * e, String * strConn)
 {
     assert(e);
-    /* check with anonymizer tables */
-    if (CBIT_TEST(Config.anonymize_headers, e->id))
-	return 0;
     /* check connection header */
     if (strConn && strListIsMember(strConn, strBuf(e->name), ','))
 	return 0;
     return 1;
 }
+
+/*
+ * Read request body contents
+ */
+void
+requestReadBody(request_t * request, char *buf, size_t size, CBCB * callback, void *cbdata)
+{
+    if (request->body_reader) {
+	if (cbdataValid(request->body_reader_data)) {
+	    request->body_reader(request, buf, size, callback, cbdata);
+	} else {
+	    debug(73, 1) ("requestReadBody: Aborted\n");
+	    request->body_reader = NULL;
+	    cbdataUnlock(request->body_reader_data);
+	    request->body_reader_data = NULL;
+	    callback(buf, 0, cbdata);	/* Signal end of body */
+	}
+    } else {
+	callback(buf, 0, cbdata);	/* Signal end of body */
+    }
+}
+
+void
+requestAbortBody(request_t * request)
+{
+    if (!request)
+	return;
+    if (request->body_reader) {
+	void *cbdata = request->body_reader_data;
+	BODY_HANDLER *handler = request->body_reader;
+	debug(73, 2) ("requestAbortBody: Aborted\n");
+	request->body_reader = NULL;
+	request->body_reader_data = NULL;
+	if (cbdataValid(cbdata))
+	    handler(request, NULL, -1, NULL, cbdata);
+	cbdataUnlock(cbdata);
+    }
+}
