@@ -1,6 +1,6 @@
 
 /*
- * $Id: delay_pools.c,v 1.17.2.2 2001/03/19 06:19:13 wessels Exp $
+ * $Id: delay_pools.c,v 1.32 2006/06/08 19:36:36 hno Exp $
  *
  * DEBUG: section 77    Delay Pools
  * AUTHOR: David Luyer <david@luyer.net>
@@ -89,7 +89,7 @@ union _delayPool {
 typedef union _delayPool delayPool;
 
 static delayPool *delay_data = NULL;
-static fd_set delay_no_delay;
+static char *delay_no_delay;
 static time_t delay_pools_last_update = 0;
 static hash_table *delay_id_ptr_hash = NULL;
 static long memory_used = 0;
@@ -134,7 +134,7 @@ void
 delayPoolsInit(void)
 {
     delay_pools_last_update = getCurrentTime();
-    FD_ZERO(&delay_no_delay);
+    delay_no_delay = xcalloc(1, Squid_MaxFD);
     cachemgrRegister("delay", "Delay Pool Levels", delayPoolStats, 0, 1);
 }
 
@@ -144,7 +144,7 @@ delayInitDelayData(unsigned short pools)
     if (!pools)
 	return;
     delay_data = xcalloc(pools, sizeof(*delay_data));
-    memory_used += sizeof(*delay_data);
+    memory_used += pools * sizeof(*delay_data);
     eventAdd("delayPoolsUpdate", delayPoolsUpdate, NULL, 1.0, 1);
     delay_id_ptr_hash = hash_create(delayIdPtrHashCmp, 256, delayIdPtrHash);
 }
@@ -160,12 +160,13 @@ delayIdZero(void *hlink)
 }
 
 void
-delayFreeDelayData(void)
+delayFreeDelayData(unsigned short pools)
 {
-    safe_free(delay_data);
-    memory_used -= sizeof(*delay_data);
     if (!delay_id_ptr_hash)
 	return;
+    eventDelete(delayPoolsUpdate, NULL);
+    safe_free(delay_data);
+    memory_used -= pools * sizeof(*delay_data);
     hashFreeItems(delay_id_ptr_hash, delayIdZero);
     hashFreeMemory(delay_id_ptr_hash);
     delay_id_ptr_hash = NULL;
@@ -283,19 +284,19 @@ delayFreeDelayPool(unsigned short pool)
 void
 delaySetNoDelay(int fd)
 {
-    FD_SET(fd, &delay_no_delay);
+    delay_no_delay[fd] = 1;
 }
 
 void
 delayClearNoDelay(int fd)
 {
-    FD_CLR(fd, &delay_no_delay);
+    delay_no_delay[fd] = 0;
 }
 
 int
 delayIsNoDelay(int fd)
 {
-    return FD_ISSET(fd, &delay_no_delay);
+    return delay_no_delay[fd];
 }
 
 static delay_id
@@ -305,19 +306,28 @@ delayId(unsigned short pool, unsigned short position)
 }
 
 delay_id
-delayClient(request_t * r)
+delayClient(clientHttpRequest * http)
 {
+    request_t *r;
     aclCheck_t ch;
     int i;
     int j;
     unsigned int host;
     unsigned short pool, position;
     unsigned char class, net;
+    assert(http);
+    r = http->request;
 
     memset(&ch, '\0', sizeof(ch));
-    ch.src_addr = r->client_addr;
+#if FOLLOW_X_FORWARDED_FOR
+    if (Config.onoff.delay_pool_uses_indirect_client) {
+	ch.src_addr = r->indirect_client_addr;
+    } else
+#endif /* FOLLOW_X_FORWARDED_FOR */
+	ch.src_addr = r->client_addr;
     ch.my_addr = r->my_addr;
     ch.my_port = r->my_port;
+    ch.conn = http->conn;
     ch.request = r;
     if (r->client_addr.s_addr == INADDR_BROADCAST) {
 	debug(77, 2) ("delayClient: WARNING: Called with 'allones' address, ignoring\n");
@@ -371,6 +381,7 @@ delayClient(request_t * r)
 	    delay_data[pool].class3->network[255] =
 		(int) (((double) Config.Delay.rates[pool]->network.max_bytes *
 		    Config.Delay.initial) / 100);
+	    delay_data[pool].class3->individual_map[i][0] = 255;
 	}
     } else {
 	for (i = 0; i < NET_MAP_SZ; i++) {
@@ -393,7 +404,6 @@ delayClient(request_t * r)
 	position |= 255;
 	if (!(delay_data[pool].class3->individual_255_used[i / 8] & (1 << (i % 8)))) {
 	    delay_data[pool].class3->individual_255_used[i / 8] |= (1 << (i % 8));
-	    assert(position < C3_IND_SZ);
 	    delay_data[pool].class3->individual[position] =
 		(int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
 		    Config.Delay.initial) / 100);
@@ -411,7 +421,6 @@ delayClient(request_t * r)
 	    assert(j < (IND_MAP_SZ - 1));
 	    delay_data[pool].class3->individual_map[i][j + 1] = 255;
 	    position |= j;
-	    assert(position < C3_IND_SZ);
 	    delay_data[pool].class3->individual[position] =
 		(int) (((double) Config.Delay.rates[pool]->individual.max_bytes *
 		    Config.Delay.initial) / 100);
@@ -488,7 +497,7 @@ delayUpdateClass3(class3DelayPool * class3, delaySpecSet * rates, int incr)
      * this loop starts at 0 or 255 and ends at 254 unless terminated earlier
      * by finding the end of the map.  note as above that 255 + 1 = 0.
      */
-    for (i = (class3->network_255_used ? 0 : 255);; ++i) {
+    for (i = (class3->network_255_used ? 255 : 0);; ++i) {
 	if (i != 255 && class3->network_map[i] == 255)
 	    return;
 	if (individual_restore_bytes != -incr) {
@@ -650,8 +659,6 @@ delayMostBytesWanted(const MemObject * mem, int max)
 	sc = (store_client *) node->data;
 	if (sc->callback_data == NULL)	/* open slot */
 	    continue;
-	if (sc->type != STORE_MEM_CLIENT)
-	    continue;
 	i = delayBytesWanted(sc->delay_id, i, max);
 	found = 1;
     }
@@ -659,7 +666,7 @@ delayMostBytesWanted(const MemObject * mem, int max)
 }
 
 delay_id
-delayMostBytesAllowed(const MemObject * mem)
+delayMostBytesAllowed(const MemObject * mem, size_t * read_sz)
 {
     int j;
     int jmax = -1;
@@ -670,14 +677,17 @@ delayMostBytesAllowed(const MemObject * mem)
 	sc = (store_client *) node->data;
 	if (sc->callback_data == NULL)	/* open slot */
 	    continue;
-	if (sc->type != STORE_MEM_CLIENT)
-	    continue;
-	j = delayBytesWanted(sc->delay_id, 0, SQUID_TCP_SO_RCVBUF);
+	j = delayBytesWanted(sc->delay_id, 0, INT_MAX);
 	if (j > jmax) {
 	    jmax = j;
 	    d = sc->delay_id;
 	}
     }
+    if (jmax >= 0 && jmax < (int) *read_sz) {
+	if (jmax == 0)
+	    jmax = 1;
+	*read_sz = (size_t) jmax;
+    }
     return d;
 }
 
