@@ -1,6 +1,6 @@
 
 /*
- * $Id: enums.h,v 1.203.2.10 2004/02/04 17:42:28 hno Exp $
+ * $Id: enums.h,v 1.233 2006/06/22 21:52:29 hno Exp $
  *
  *
  * SQUID Web Proxy Cache          http://www.squid-cache.org/
@@ -93,6 +93,7 @@ typedef enum {
     ERR_ONLY_IF_CACHED_MISS,	/* failure to satisfy only-if-cached request */
     ERR_TOO_BIG,
     TCP_RESET,
+    ERR_INVALID_RESP,
     ERR_MAX
 } err_type;
 
@@ -114,6 +115,7 @@ typedef enum {
     ACL_IDENT,
     ACL_IDENT_REGEX,
 #endif
+    ACL_TYPE,
     ACL_PROTO,
     ACL_METHOD,
     ACL_BROWSER,
@@ -134,9 +136,18 @@ typedef enum {
     ACL_MAXCONN,
     ACL_REQ_MIME_TYPE,
     ACL_REP_MIME_TYPE,
+    ACL_REP_HEADER,
+    ACL_REQ_HEADER,
     ACL_MAX_USER_IP,
     ACL_EXTERNAL,
     ACL_URLLOGIN,
+#if USE_SSL
+    ACL_USER_CERT,
+    ACL_CA_CERT,
+#endif
+    ACL_URLGROUP,
+    ACL_EXTUSER,
+    ACL_EXTUSER_REGEX,
     ACL_ENUM_MAX
 } squid_acl;
 
@@ -177,7 +188,8 @@ typedef enum {
 
 /* recognized or "known" header fields; @?@ add more! */
 typedef enum {
-    HDR_ACCEPT,
+    HDR_UNKNOWN = -1,
+    HDR_ACCEPT = 0,
     HDR_ACCEPT_CHARSET,
     HDR_ACCEPT_ENCODING,
     HDR_ACCEPT_LANGUAGE,
@@ -188,6 +200,7 @@ typedef enum {
     HDR_CACHE_CONTROL,
     HDR_CONNECTION,
     HDR_CONTENT_BASE,
+    HDR_CONTENT_DISPOSITION,
     HDR_CONTENT_ENCODING,
     HDR_CONTENT_LANGUAGE,
     HDR_CONTENT_LENGTH,
@@ -240,6 +253,10 @@ typedef enum {
 #if X_ACCELERATOR_VARY
     HDR_X_ACCELERATOR_VARY,
 #endif
+    HDR_X_ERROR_URL,		/* errormap, requested URL */
+    HDR_X_ERROR_STATUS,		/* errormap, received HTTP status line */
+    HDR_FRONT_END_HTTPS,
+    HDR_PROXY_SUPPORT,
     HDR_OTHER,
     HDR_ENUM_END
 } http_hdr_type;
@@ -270,7 +287,8 @@ typedef enum {
     ftPCc,
     ftPContRange,
     ftPRange,
-    ftDate_1123_or_ETag
+    ftDate_1123_or_ETag,
+    ftSize
 } field_type;
 
 /* possible owners of http header */
@@ -285,7 +303,7 @@ typedef enum {
 
 typedef enum {
     HIER_NONE,
-    DIRECT,
+    HIER_DIRECT,
     SIBLING_HIT,
     PARENT_HIT,
     DEFAULT_PARENT,
@@ -306,6 +324,9 @@ typedef enum {
     CARP,
 #endif
     ANY_OLD_PARENT,
+    USERHASH_PARENT,
+    SOURCEHASH_PARENT,
+    PINNED,
     HIER_MAX
 } hier_code;
 
@@ -392,6 +413,8 @@ enum {
     METHOD_SUBSCRIBE,
     METHOD_UNSUBSCRIBE,
     METHOD_POLL,
+    METHOD_REPORT,
+    /* Extension methods must be last, Add any new methods before this line */
     METHOD_EXT00,
     METHOD_EXT01,
     METHOD_EXT02,
@@ -480,7 +503,8 @@ typedef enum {
     HTTP_GATEWAY_TIMEOUT = 504,
     HTTP_HTTP_VERSION_NOT_SUPPORTED = 505,
     HTTP_INSUFFICIENT_STORAGE = 507,	/* RFC2518 section 10.6 */
-    HTTP_INVALID_HEADER = 600	/* Squid header parsing error */
+    HTTP_INVALID_HEADER = 600,	/* Squid header parsing error */
+    HTTP_HEADER_TOO_LARGE = 601	/* Header too large to process */
 } http_status;
 
 /*
@@ -502,7 +526,8 @@ enum {
     ENTRY_NEGCACHED,
     ENTRY_VALIDATED,
     ENTRY_BAD_LENGTH,
-    ENTRY_ABORTED
+    ENTRY_ABORTED,
+    ENTRY_DEFER_READ
 #if UNUSED_CODE
     ENTRY_DONT_LOG
 #endif
@@ -526,6 +551,7 @@ typedef enum {
     AUTH_BASIC,
     AUTH_NTLM,
     AUTH_DIGEST,
+    AUTH_NEGOTIATE,
     AUTH_BROKEN			/* known type, but broken data */
 } auth_type_t;
 
@@ -559,6 +585,7 @@ typedef enum {
     MEM_ACL_IP_DATA,
     MEM_ACL_LIST,
     MEM_ACL_NAME_LIST,
+    MEM_ACL_REQUEST_TYPE,
     MEM_AUTH_USER_T,
     MEM_AUTH_USER_HASH,
     MEM_ACL_PROXY_AUTH_MATCH,
@@ -601,7 +628,9 @@ typedef enum {
     MEM_EVENT,
     MEM_TLV,
     MEM_SWAP_LOG_DATA,
-    MEM_CLIENT_REQ_BUF,
+#if USE_SSL
+    MEM_ACL_CERT_DATA,
+#endif
     MEM_MAX
 } mem_type;
 
@@ -618,6 +647,8 @@ enum {
     STORE_META_HITMETERING,	/* reserved for hit metering */
     STORE_META_VALID,
     STORE_META_VARY_HEADERS,	/* Stores Vary request headers */
+    STORE_META_STD_LFS,		/* standard metadata in lfs format */
+    STORE_META_OBJSIZE,		/* object size, if its known */
     STORE_META_END
 };
 
@@ -633,6 +664,7 @@ typedef enum {
     SWAP_LOG_NOP,
     SWAP_LOG_ADD,
     SWAP_LOG_DEL,
+    SWAP_LOG_VERSION,
     SWAP_LOG_MAX
 } swap_log_op;
 
@@ -712,8 +744,8 @@ enum {
     VARY_CANCEL
 };
 
-/* CygWin & Windows NT Port */
-#if defined(_SQUID_MSWIN_) || defined(_SQUID_CYGWIN_)
+/* Windows Port */
+#ifdef _SQUID_WIN32_
 /*
  * Supported Windows OS types codes
  */
@@ -726,9 +758,27 @@ enum {
     _WIN_OS_WINNT,
     _WIN_OS_WIN2K,
     _WIN_OS_WINXP,
-    _WIN_OS_WINNET
+    _WIN_OS_WINNET,
+    _WIN_OS_WINLON
 };
 
 #endif
 
+/*
+ * Special case pending filedescriptors. Set in fd_table[fd].read/write_pending
+ */
+typedef enum {
+    COMM_PENDING_NORMAL,	/* No special processing required */
+    COMM_PENDING_WANTS_READ,	/* need to read, no matter what commSetSelect indicates */
+    COMM_PENDING_WANTS_WRITE,	/* need to write, no matter what commSetSelect indicates */
+    COMM_PENDING_NOW		/* needs to be called again, without needing to wait for readiness
+				 * for example when data is already buffered etc */
+} comm_pending;
+
+typedef enum {
+    ST_OP_NONE,
+    ST_OP_OPEN,
+    ST_OP_CREATE
+} store_op_t;
+
 #endif /* SQUID_ENUMS_H */
