@@ -1,6 +1,6 @@
 
 /*
- * $Id: store.c,v 1.544.2.8 2005/03/26 02:50:53 hno Exp $
+ * $Id: store.c,v 1.561 2006/06/30 15:05:38 hno Exp $
  *
  * DEBUG: section 20    Storage Manager
  * AUTHOR: Harvest Derived
@@ -121,6 +121,7 @@ new_MemObject(const char *url, const char *log_url)
 #endif
     mem->log_url = xstrdup(log_url);
     mem->object_sz = -1;
+    mem->serverfd = -1;
     /* XXX account log_url */
     debug(20, 3) ("new_MemObject: returning %p\n", mem);
     return mem;
@@ -161,6 +162,10 @@ destroy_MemObject(StoreEntry * e)
      */
     assert(mem->clients.head == NULL);
 #endif
+    if (mem->ims_entry) {
+	storeUnlockObject(mem->ims_entry);
+	mem->ims_entry = NULL;
+    }
     httpReplyDestroy(mem->reply);
     requestUnlink(mem->request);
     mem->request = NULL;
@@ -168,6 +173,7 @@ destroy_MemObject(StoreEntry * e)
     safe_free(mem->url);
     safe_free(mem->log_url);	/* XXX account log_url */
     safe_free(mem->vary_headers);
+    safe_free(mem->vary_encoding);
     memFree(mem, MEM_MEMOBJECT);
 }
 
@@ -330,6 +336,13 @@ storeGetPublic(const char *uri, const method_t method)
 StoreEntry *
 storeGetPublicByRequestMethod(request_t * req, const method_t method)
 {
+    if (req->vary) {
+	/* Varying objects... */
+	if (req->vary->key)
+	    return storeGet(storeKeyScan(req->vary->key));
+	else
+	    return NULL;
+    }
     return storeGet(storeKeyPublicByRequestMethod(req, method));
 }
 
@@ -375,6 +388,560 @@ storeSetPrivateKey(StoreEntry * e)
     storeHashInsert(e, newkey);
 }
 
+typedef struct {
+    StoreEntry *oe;
+    StoreEntry *e;
+    store_client *sc;
+    char *url;
+    char *key;
+    char *vary_headers;
+    char *accept_encoding;
+    char *etag;
+    squid_off_t seen_offset;
+    char *buf;
+    size_t buf_size;
+    size_t buf_offset;
+    int done:1;
+    struct {
+	char *key;
+	char *etag;
+	char *accept_encoding;
+	int this_key:1;
+	int key_used:1;
+	int ignore:1;
+    } current;
+} AddVaryState;
+
+CBDATA_TYPE(AddVaryState);
+static void
+free_AddVaryState(void *data)
+{
+    AddVaryState *state = data;
+    debug(11, 2) ("free_AddVaryState: %p\n", data);
+    if (!state->done && state->key) {
+	storeAppendPrintf(state->e, "Key: %s\n", state->key);
+	if (state->accept_encoding)
+	    storeAppendPrintf(state->e, "Accept-Encoding: %s\n", state->accept_encoding);
+	if (state->etag)
+	    storeAppendPrintf(state->e, "ETag: %s\n", state->etag);
+	storeAppendPrintf(state->e, "VaryData: %s\n", state->vary_headers);
+    }
+    storeBufferFlush(state->e);
+    storeTimestampsSet(state->e);
+    storeComplete(state->e);
+    storeTimestampsSet(state->e);
+    storeUnlockObject(state->e);
+    state->e = NULL;
+    if (state->sc) {
+	storeClientUnregister(state->sc, state->oe, state);
+	state->sc = NULL;
+    }
+    if (state->oe) {
+	storeUnlockObject(state->oe);
+	state->oe = NULL;
+    }
+    safe_free(state->url);
+    safe_free(state->key);
+    safe_free(state->vary_headers);
+    safe_free(state->accept_encoding);
+    safe_free(state->etag);
+    safe_free(state->current.key);
+    safe_free(state->current.etag);
+    safe_free(state->current.accept_encoding);
+    if (state->buf) {
+	memFreeBuf(state->buf_size, state->buf);
+	state->buf = NULL;
+    }
+}
+
+static int inline
+strmatchbeg(const char *search, const char *match, int maxlen)
+{
+    int mlen = strlen(match);
+    if (maxlen < mlen)
+	return -1;
+    return strncmp(search, match, mlen);
+}
+
+static int inline
+strmatch(const char *search, const char *match, int maxlen)
+{
+    int mlen = strlen(match);
+    if (maxlen < mlen)
+	return -1;
+    return strncmp(search, match, maxlen);
+}
+
+static void
+storeAddVaryFlush(AddVaryState * state)
+{
+    if (state->current.ignore || state->current.key_used) {
+	/* do nothing */
+    } else if (state->current.this_key) {
+	if (state->current.key)
+	    storeAppendPrintf(state->e, "Key: %s\n", state->current.key);
+	else
+	    storeAppendPrintf(state->e, "Key: %s\n", state->key);
+	if (state->accept_encoding)
+	    storeAppendPrintf(state->e, "Accept-Encoding: %s\n", state->accept_encoding);
+	if (state->etag)
+	    storeAppendPrintf(state->e, "ETag: %s\n", state->etag);
+	storeAppendPrintf(state->e, "VaryData: %s\n", state->vary_headers);
+	state->done = 1;
+	state->current.key_used = 1;
+    } else if (state->current.key) {
+	storeAppendPrintf(state->e, "Key: %s\n", state->current.key);
+	safe_free(state->current.key);
+	if (state->current.accept_encoding)
+	    storeAppendPrintf(state->e, "Accept-Encoding: %s\n", state->current.accept_encoding);
+	if (state->current.etag) {
+	    storeAppendPrintf(state->e, "ETag: %s\n", state->current.etag);
+	    safe_free(state->current.etag);
+	}
+	state->current.key_used = 1;
+    }
+}
+
+static int
+strcmpnull(const char *a, const char *b)
+{
+    if (a && b)
+	return strcmp(a, b);
+    else if (a)
+	return 1;
+    else if (b)
+	return -1;
+    return 0;
+}
+
+static int
+strncmpnull(const char *a, const char *b, size_t n)
+{
+    if (a && b)
+	return strncmp(a, b, n);
+    else if (a)
+	return 1;
+    else if (b)
+	return -1;
+    return 0;
+}
+
+static void
+storeAddVaryReadOld(void *data, char *buf, ssize_t size)
+{
+    AddVaryState *state = data;
+    size_t l = size + state->buf_offset;
+    char *e;
+    char *p = state->buf;
+    debug(11, 3) ("storeAddVaryReadOld: %p seen_offset=%" PRINTF_OFF_T " buf_offset=%d size=%d\n", data, state->seen_offset, (int) state->buf_offset, (int) size);
+    if (size <= 0) {
+	debug(11, 2) ("storeAddVaryReadOld: DONE\n");
+	cbdataFree(state);
+	return;
+    }
+    if (state->seen_offset != 0) {
+	state->seen_offset = state->seen_offset + size;
+    } else {
+	int hdr_sz;
+	if (!state->oe->mem_obj->reply)
+	    goto invalid_marker_obj;
+	if (!strLen(state->oe->mem_obj->reply->content_type))
+	    goto invalid_marker_obj;
+	if (strCmp(state->oe->mem_obj->reply->content_type, "x-squid-internal/vary") != 0) {
+	  invalid_marker_obj:
+	    debug(11, 2) ("storeAddVaryReadOld: %p (%s) is not a Vary maker object, ignoring\n", data, storeUrl(state->oe));
+	    cbdataFree(state);
+	    return;
+	}
+	hdr_sz = state->e->mem_obj->reply->hdr_sz;
+	state->seen_offset = state->e->mem_obj->reply->hdr_sz;
+	if (l >= state->e->mem_obj->reply->hdr_sz) {
+	    state->seen_offset = l;
+	    l -= hdr_sz;
+	    p += hdr_sz;
+	} else {
+	    l = 0;
+	    state->seen_offset = hdr_sz;
+	}
+    }
+    while (l && (e = memchr(p, '\n', l)) != NULL) {
+	int l2;
+	char *p2;
+	if (strmatchbeg(p, "Key: ", l) == 0) {
+	    /* key field */
+	    p2 = p + 5;
+	    l2 = e - p2;
+	    if (state->current.this_key) {
+		storeAddVaryFlush(state);
+	    }
+	    safe_free(state->current.key);
+	    safe_free(state->current.etag);
+	    safe_free(state->current.accept_encoding);
+	    memset(&state->current, 0, sizeof(state->current));
+	    state->current.key = xmalloc(l2 + 1);
+	    memcpy(state->current.key, p2, l2);
+	    state->current.key[l2] = '\0';
+	    if (state->key) {
+		if (strcmp(state->current.key, state->key) == 0)
+		    state->current.this_key = 1;
+	    }
+	    debug(11, 3) ("storeAddVaryReadOld: Key: %s%s\n", state->current.key, state->current.this_key ? " (THIS)" : "");
+	} else if (strmatchbeg(p, "ETag: ", l) == 0) {
+	    /* etag field */
+	    p2 = p + 6;
+	    l2 = e - p2;
+	    safe_free(state->current.etag);
+	    state->current.etag = xmalloc(l2 + 1);
+	    memcpy(state->current.etag, p2, l2);
+	    state->current.etag[l2] = '\0';
+	    if (state->etag && strcmp(state->current.etag, state->etag) == 0) {
+		if (state->accept_encoding && strcmpnull(state->accept_encoding, state->current.accept_encoding) != 0) {
+		    /* Skip this match. It's not ours */
+		} else if (!state->key) {
+		    state->current.this_key = 1;
+		} else if (!state->current.this_key) {
+		    const cache_key *oldkey = storeKeyScan(state->current.key);
+		    StoreEntry *old_e = storeGet(oldkey);
+		    if (old_e)
+			storeRelease(old_e);
+		    if (!state->done) {
+			safe_free(state->current.key);
+			state->current.key = xstrdup(state->key);
+			state->current.this_key = 1;
+		    } else {
+			state->current.ignore = 1;
+		    }
+		}
+	    } else if (state->current.this_key) {
+		state->current.ignore = 1;
+	    }
+	    debug(11, 2) ("storeAddVaryReadOld: ETag: %s%s%s\n", state->current.etag, state->current.this_key ? " (THIS)" : "", state->current.ignore ? " (IGNORE)" : "");
+	} else if (!state->current.ignore && strmatchbeg(p, "VaryData: ", l) == 0) {
+	    /* vary field */
+	    p2 = p + 10;
+	    l2 = e - p2;
+	    storeAddVaryFlush(state);
+	    if (strmatch(p2, state->vary_headers, l2) != 0) {
+		storeAppend(state->e, p, e - p + 1);
+		debug(11, 3) ("storeAddVaryReadOld: %s\n", p);
+	    }
+	} else if (strmatchbeg(p, "Accept-Encoding: ", l) == 0) {
+	    p2 = p + 17;
+	    l2 = e - p2;
+	    safe_free(state->current.accept_encoding);
+	    state->current.accept_encoding = xmalloc(l2 + 1);
+	    memcpy(state->current.accept_encoding, p2, l2);
+	    state->current.accept_encoding[l2] = '\0';
+	}
+	e += 1;
+	l -= e - p;
+	p = e;
+	if (l == 0)
+	    break;
+	assert(p <= (buf + size));
+    }
+    if (p == state->buf && size == state->buf_size) {
+	/* Oops.. the buffer size is not sufficient. Grow */
+	if (state->buf_size < 65536) {
+	    debug(11, 2) ("storeAddVaryReadOld: Increasing entry buffer size to %d\n", (int) state->buf_size * 2);
+	    state->buf = memReallocBuf(state->buf, state->buf_size * 2, &state->buf_size);
+	} else {
+	    /* This does not look good. Bail out. This should match the size <= 0 case above */
+	    debug(11, 1) ("storeAddVaryReadOld: Buffer very large and still can't fit the data.. bailing out\n");
+	    cbdataFree(state);
+	    return;
+	}
+    }
+    state->buf_offset = l;
+    if (l)
+	memmove(state->buf, p, l);
+    debug(11, 3) ("storeAddVaryReadOld: %p seen_offset=%" PRINTF_OFF_T " buf_offset=%d\n", data, state->seen_offset, (int) state->buf_offset);
+    storeClientCopy(state->sc, state->oe,
+	state->seen_offset,
+	state->seen_offset,
+	state->buf_size - state->buf_offset,
+	state->buf + state->buf_offset,
+	storeAddVaryReadOld,
+	state);
+}
+
+/*
+ * Adds/updates a Vary record.
+ * For updates only one of key or etag needs to be specified
+ * At leas one of key or etag must be specified, preferably both.
+ */
+void
+storeAddVary(const char *url, const char *log_url, const method_t method, const cache_key * key, const char *etag, const char *vary, const char *vary_headers, const char *accept_encoding)
+{
+    AddVaryState *state;
+    http_version_t version;
+    request_flags flags = null_request_flags;
+    CBDATA_INIT_TYPE_FREECB(AddVaryState, free_AddVaryState);
+    state = cbdataAlloc(AddVaryState);
+    state->url = xstrdup(url);
+    if (key)
+	state->key = xstrdup(storeKeyText(key));
+    state->vary_headers = xstrdup(vary_headers);
+    if (accept_encoding)
+	state->accept_encoding = xstrdup(accept_encoding);
+    if (etag)
+	state->etag = xstrdup(etag);
+    state->oe = storeGetPublic(url, method);
+    debug(11, 2) ("storeAddVary: %s (%s) %s %s\n",
+	state->url, state->key, state->vary_headers, state->etag);
+    if (state->oe)
+	storeLockObject(state->oe);
+    flags.cachable = 1;
+    state->e = storeCreateEntry(url, log_url, flags, method);
+    httpBuildVersion(&version, 1, 0);
+    httpReplySetHeaders(state->e->mem_obj->reply, version, HTTP_OK, "Internal marker object", "x-squid-internal/vary", -1, -1, squid_curtime + 100000);
+    httpHeaderPutStr(&state->e->mem_obj->reply->header, HDR_VARY, vary);
+    storeSetPublicKey(state->e);
+    storeBuffer(state->e);
+    httpReplySwapOut(state->e->mem_obj->reply, state->e);
+    storeBufferFlush(state->e);
+    if (state->oe) {
+	/* Here we need to tack on the old etag/vary information, and we should
+	 * merge, clean up etc
+	 *
+	 * Suggestion:
+	 * swap in the old file, looking for ETag, Key and VaryData. If a match is
+	 * found then 
+	 * - on ETag, update the key, and expire the old object if different
+	 * - on Key, drop the old data if ETag is different, else nothing
+	 * - on VaryData, remove the line if a different key. If this makes
+	 *   the searched key "empty" then expire it and remove it from the
+	 *   map
+	 * - VaryData is added last in the Key record it corresponds to (after
+	 *   modifications above)
+	 */
+	/* Swap in the dummy Vary object */
+	if (!state->oe->mem_obj) {
+	    storeCreateMemObject(state->oe, state->url, log_url);
+	    state->oe->mem_obj->method = method;
+	}
+	state->sc = storeClientRegister(state->oe, state);
+	state->buf = memAllocBuf(4096, &state->buf_size);
+	debug(11, 3) ("storeAddVary: %p\n", state);
+	storeClientCopy(state->sc, state->oe, 0, 0,
+	    state->buf_size,
+	    state->buf,
+	    storeAddVaryReadOld,
+	    state);
+	return;
+    } else {
+	cbdataFree(state);
+    }
+}
+
+static MemPool *VaryData_pool = NULL;
+
+void
+storeLocateVaryDone(VaryData * data)
+{
+    int i;
+    safe_free(data->key);
+    data->etag = NULL;		/* points to an entry in etags */
+    for (i = 0; i < data->etags.count; i++) {
+	safe_free(data->etags.items[i]);
+    }
+    arrayClean(&data->etags);
+    memPoolFree(VaryData_pool, data);
+}
+
+typedef struct {
+    VaryData *data;
+    STLVCB *callback;
+    void *callback_data;
+    StoreEntry *e;
+    store_client *sc;
+    char *buf;
+    size_t buf_size;
+    size_t buf_offset;
+    char *vary_data;
+    char *accept_encoding;
+    squid_off_t seen_offset;
+    struct {
+	int ignore;
+	int encoding_ok;
+	char *key;
+	char *etag;
+    } current;
+} LocateVaryState;
+
+CBDATA_TYPE(LocateVaryState);
+
+static void
+storeLocateVaryCallback(LocateVaryState * state)
+{
+    if (cbdataValid(state->callback_data)) {
+	VaryData *data = state->data;
+	if (data->key || data->etags.count) {
+	    state->callback(data, state->callback_data);
+	    state->data = NULL;	/* now owned by the caller */
+	} else {
+	    state->callback(NULL, state->callback_data);
+	}
+    }
+    cbdataUnlock(state->callback_data);
+    if (state->data) {
+	storeLocateVaryDone(state->data);
+	state->data = NULL;
+    }
+    state->current.etag = NULL;	/* shared by data->entries[x] */
+    safe_free(state->vary_data);
+    safe_free(state->accept_encoding);
+    safe_free(state->current.key);
+    if (state->sc) {
+	storeClientUnregister(state->sc, state->e, state);
+	state->sc = NULL;
+    }
+    if (state->e) {
+	storeUnlockObject(state->e);
+	state->e = NULL;
+    }
+    if (state->buf) {
+	memFreeBuf(state->buf_size, state->buf);
+	state->buf = NULL;
+    }
+    cbdataFree(state);
+    debug(11, 2) ("storeLocateVary: DONE\n");
+}
+
+static void
+storeLocateVaryRead(void *data, char *buf, ssize_t size)
+{
+    LocateVaryState *state = data;
+    char *e;
+    char *p = state->buf;
+    size_t l = size + state->buf_offset;
+    debug(11, 3) ("storeLocateVaryRead: %s %p seen_offset=%" PRINTF_OFF_T " buf_offset=%d size=%d\n", state->vary_data, data, state->seen_offset, (int) state->buf_offset, (int) size);
+    if (size <= 0) {
+	storeLocateVaryCallback(state);
+	return;
+    }
+    state->seen_offset = state->seen_offset + size;
+    while ((e = memchr(p, '\n', l)) != NULL) {
+	int l2;
+	char *p2;
+	if (strmatchbeg(p, "Key: ", l) == 0) {
+	    /* key field */
+	    p2 = p + 5;
+	    l2 = e - p2;
+	    safe_free(state->current.key);
+	    state->current.etag = NULL;		/* saved in data.etags[] */
+	    state->current.ignore = 0;
+	    state->current.encoding_ok = !state->accept_encoding;
+	    state->current.key = xmalloc(l2 + 1);
+	    memcpy(state->current.key, p2, l2);
+	    state->current.key[l2] = '\0';
+	    debug(11, 3) ("storeLocateVaryRead: Key: %s\n", state->current.key);
+	} else if (state->current.ignore) {
+	    /* Skip this entry */
+	} else if (strmatchbeg(p, "ETag: ", l) == 0) {
+	    /* etag field */
+	    char *etag;
+	    if (state->current.encoding_ok) {
+		p2 = p + 6;
+		l2 = e - p2;
+		etag = xmalloc(l2 + 1);
+		memcpy(etag, p2, l2);
+		etag[l2] = '\0';
+		state->current.etag = etag;
+		arrayAppend(&state->data->etags, etag);
+		debug(11, 3) ("storeLocateVaryRead: ETag: %s\n", etag);
+	    } else {
+		state->current.ignore = 1;
+	    }
+	} else if (strmatchbeg(p, "VaryData: ", l) == 0) {
+	    /* vary field */
+	    p2 = p + 10;
+	    l2 = e - p2;
+	    if (strmatch(p2, state->vary_data, l2) == 0) {
+		/* A matching vary header found */
+		safe_free(state->data->key);
+		state->data->key = xstrdup(state->current.key);
+		state->data->etag = state->current.etag;
+		debug(11, 2) ("storeLocateVaryRead: MATCH! %s %s\n", state->current.key, state->current.etag);
+	    }
+	} else if (strmatchbeg(p, "Accept-Encoding: ", l) == 0) {
+	    p2 = p + 17;
+	    l2 = e - p2;
+	    if (strncmpnull(state->accept_encoding, p2, l2) == 0 && state->accept_encoding[l2] == '\0') {
+		state->current.encoding_ok = 1;
+	    }
+	}
+	e += 1;
+	l -= e - p;
+	p = e;
+	if (l == 0)
+	    break;
+	assert(l > 0);
+	assert(p < (buf + size));
+    }
+    state->buf_offset = l;
+    if (l)
+	memmove(state->buf, p, l);
+    if (state->buf_offset == state->buf_size) {
+	/* Oops.. the buffer size is not sufficient. Grow */
+	if (state->buf_size < 65536) {
+	    debug(11, 2) ("storeLocateVaryRead: Increasing entry buffer size to %d\n", (int) state->buf_size * 2);
+	    state->buf = memReallocBuf(state->buf, state->buf_size * 2, &state->buf_size);
+	} else {
+	    /* This does not look good. Bail out. This should match the size <= 0 case above */
+	    debug(11, 1) ("storeLocateVaryRead: Buffer very large and still can't fit the data.. bailing out\n");
+	    storeLocateVaryCallback(state);
+	    return;
+	}
+    }
+    debug(11, 3) ("storeLocateVaryRead: %p seen_offset=%" PRINTF_OFF_T " buf_offset=%d\n", data, state->seen_offset, (int) state->buf_offset);
+    storeClientCopy(state->sc, state->e,
+	state->seen_offset,
+	state->seen_offset,
+	state->buf_size - state->buf_offset,
+	state->buf + state->buf_offset,
+	storeLocateVaryRead,
+	state);
+}
+
+void
+storeLocateVary(StoreEntry * e, int offset, const char *vary_data, String accept_encoding, STLVCB * callback, void *cbdata)
+{
+    LocateVaryState *state;
+    debug(11, 2) ("storeLocateVary: %s\n", vary_data);
+    CBDATA_INIT_TYPE(LocateVaryState);
+    if (!VaryData_pool)
+	VaryData_pool = memPoolCreate("VaryData", sizeof(VaryData));
+    state = cbdataAlloc(LocateVaryState);
+    state->vary_data = xstrdup(vary_data);
+    if (strBuf(accept_encoding))
+	state->accept_encoding = xstrdup(strBuf(accept_encoding));
+    state->data = memPoolAlloc(VaryData_pool);
+    stringClean(&accept_encoding);
+    state->e = e;
+    storeLockObject(state->e);
+    state->callback_data = cbdata;
+    cbdataLock(cbdata);
+    state->callback = callback;
+    state->buf = memAllocBuf(4096, &state->buf_size);
+    state->sc = storeClientRegister(state->e, state);
+    state->seen_offset = offset;
+    if (strCmp(e->mem_obj->reply->content_type, "x-squid-internal/vary") != 0) {
+	/* This is not our Vary marker object. Bail out. */
+	debug(33, 1) ("storeLocateVary: Not our vary marker object, %s = '%s', '%s'/'%s'\n",
+	    storeKeyText(e->hash.key), e->mem_obj->url, vary_data, strBuf(accept_encoding) ? strBuf(accept_encoding) : "-");
+	storeLocateVaryCallback(state);
+	return;
+    }
+    storeClientCopy(state->sc, state->e,
+	state->seen_offset,
+	state->seen_offset,
+	state->buf_size,
+	state->buf,
+	storeLocateVaryRead,
+	state);
+}
+
 void
 storeSetPublicKey(StoreEntry * e)
 {
@@ -423,36 +990,27 @@ storeSetPublicKey(StoreEntry * e)
 		    request->vary_headers = xstrdup(vary);
 	    }
 	}
-	if (mem->vary_headers && !storeGetPublic(mem->url, mem->method)) {
-	    /* Create "vary" base object */
-	    http_version_t version;
-	    String vary;
-	    pe = storeCreateEntry(mem->url, mem->log_url, request->flags, request->method);
-	    httpBuildVersion(&version, 1, 0);
-	    httpReplySetHeaders(pe->mem_obj->reply, version, HTTP_OK, "Internal marker object", "x-squid-internal/vary", -1, -1, squid_curtime + 100000);
-	    vary = httpHeaderGetList(&mem->reply->header, HDR_VARY);
-	    if (strBuf(vary)) {
-		httpHeaderPutStr(&pe->mem_obj->reply->header, HDR_VARY, strBuf(vary));
-		stringClean(&vary);
-	    }
-#if X_ACCELERATOR_VARY
-	    vary = httpHeaderGetList(&mem->reply->header, HDR_X_ACCELERATOR_VARY);
-	    if (strBuf(vary)) {
-		httpHeaderPutStr(&pe->mem_obj->reply->header, HDR_X_ACCELERATOR_VARY, strBuf(vary));
-		stringClean(&vary);
-	    }
-#endif
-	    storeSetPublicKey(pe);
-	    storeBuffer(pe);
-	    httpReplySwapOut(pe->mem_obj->reply, pe);
-	    storeBufferFlush(pe);
-	    storeTimestampsSet(pe);
-	    storeComplete(pe);
-	    storeUnlockObject(pe);
-	}
 	newkey = storeKeyPublicByRequest(mem->request);
-    } else
+	if (mem->vary_headers) {
+	    String vary = StringNull;
+	    String varyhdr;
+	    varyhdr = httpHeaderGetList(&mem->reply->header, HDR_VARY);
+	    if (strBuf(varyhdr))
+		strListAdd(&vary, strBuf(varyhdr), ',');
+	    stringClean(&varyhdr);
+#if X_ACCELERATOR_VARY
+	    /* This needs to match the order in http.c:httpMakeVaryMark */
+	    varyhdr = httpHeaderGetList(&mem->reply->header, HDR_X_ACCELERATOR_VARY);
+	    if (strBuf(varyhdr))
+		strListAdd(&vary, strBuf(varyhdr), ',');
+	    stringClean(&varyhdr);
+#endif
+	    storeAddVary(mem->url, mem->log_url, mem->method, newkey, httpHeaderGetStr(&mem->reply->header, HDR_ETAG), strBuf(vary), mem->vary_headers, mem->vary_encoding);
+	    stringClean(&vary);
+	}
+    } else {
 	newkey = storeKeyPublic(mem->url, mem->method);
+    }
     if ((e2 = (StoreEntry *) hash_lookup(store_table, newkey))) {
 	debug(20, 3) ("storeSetPublicKey: Making old '%s' private.\n", mem->url);
 	storeSetPrivateKey(e2);
@@ -521,6 +1079,7 @@ storeAppend(StoreEntry * e, const char *buf, int len)
     assert(mem != NULL);
     assert(len >= 0);
     assert(e->store_status == STORE_PENDING);
+    mem->refresh_timestamp = squid_curtime;
     if (len) {
 	debug(20, 5) ("storeAppend: appending %d bytes for '%s'\n",
 	    len,
@@ -1066,6 +1625,10 @@ storeEntryValidToSend(StoreEntry * e)
 	    return 0;
     if (EBIT_TEST(e->flags, ENTRY_ABORTED))
 	return 0;
+    /* Entries which seem to have got stuck is not valid to send to new clients */
+    if (e->store_status == STORE_PENDING)
+	if (!e->mem_obj || e->mem_obj->refresh_timestamp + 30 < squid_curtime)
+	    return 0;
     return 1;
 }
 
@@ -1111,7 +1674,7 @@ storeRegisterAbort(StoreEntry * e, STABH * cb, void *data)
 }
 
 void
-storeUnregisterAbort(StoreEntry * e)
+storeClientUnregisterAbort(StoreEntry * e)
 {
     MemObject *mem = e->mem_obj;
     assert(mem);
@@ -1232,9 +1795,11 @@ storeBuffer(StoreEntry * e)
 void
 storeBufferFlush(StoreEntry * e)
 {
-    EBIT_CLR(e->flags, DELAY_SENDING);
-    InvokeHandlers(e);
-    storeSwapOut(e);
+    if (EBIT_TEST(e->flags, DELAY_SENDING)) {
+	EBIT_CLR(e->flags, DELAY_SENDING);
+	InvokeHandlers(e);
+	storeSwapOut(e);
+    }
 }
 
 squid_off_t
@@ -1378,3 +1943,36 @@ storeSwapFileNumberSet(StoreEntry * e, sfileno filn)
     }
 }
 #endif
+
+/* Defer server-side reads */
+void
+storeDeferRead(StoreEntry * e, int fd)
+{
+    MemObject *mem = e->mem_obj;
+    EBIT_SET(e->flags, ENTRY_DEFER_READ);
+    if (fd >= 0) {
+	mem->serverfd = fd;
+	commDeferFD(fd);
+    }
+}
+
+/* Resume reading from the server-side */
+void
+storeResumeRead(StoreEntry * e)
+{
+    MemObject *mem = e->mem_obj;
+    EBIT_CLR(e->flags, ENTRY_DEFER_READ);
+    if (mem->serverfd != -1) {
+	commResumeFD(mem->serverfd);
+	mem->serverfd = -1;
+    }
+}
+
+/* Reset defer state when FD goes away under our feets */
+void
+storeResetDefer(StoreEntry * e)
+{
+    EBIT_CLR(e->flags, ENTRY_DEFER_READ);
+    if (e->mem_obj)
+	e->mem_obj->serverfd = -1;
+}
