@@ -1,6 +1,6 @@
 
 /*
- * $Id: store_swapmeta.c,v 1.17 2001/10/24 08:52:37 hno Exp $
+ * $Id: store_swapmeta.c,v 1.20 2006/05/20 13:05:58 hno Exp $
  *
  * DEBUG: section 20    Storage Manager Swapfile Metadata
  * AUTHOR: Kostas Anagnostakis
@@ -68,13 +68,21 @@ storeSwapMetaBuild(StoreEntry * e)
     tlv **T = &TLV;
     const char *url;
     const char *vary;
+    const squid_off_t objsize = objectLen(e);
     assert(e->mem_obj != NULL);
     assert(e->swap_status == SWAPOUT_WRITING);
     url = storeUrl(e);
     debug(20, 3) ("storeSwapMetaBuild: %s\n", url);
     T = storeSwapTLVAdd(STORE_META_KEY, e->hash.key, MD5_DIGEST_CHARS, T);
+#if SIZEOF_SQUID_FILE_SZ == SIZEOF_SIZE_T
     T = storeSwapTLVAdd(STORE_META_STD, &e->timestamp, STORE_HDR_METASIZE, T);
+#else
+    T = storeSwapTLVAdd(STORE_META_STD_LFS, &e->timestamp, STORE_HDR_METASIZE, T);
+#endif
     T = storeSwapTLVAdd(STORE_META_URL, url, strlen(url) + 1, T);
+    if (objsize > -1) {
+	T = storeSwapTLVAdd(STORE_META_OBJSIZE, &objsize, sizeof(objsize), T);
+    }
     vary = e->mem_obj->vary_headers;
     if (vary)
 	T = storeSwapTLVAdd(STORE_META_VARY_HEADERS, vary, strlen(vary) + 1, T);
@@ -86,14 +94,13 @@ storeSwapMetaPack(tlv * tlv_list, int *length)
 {
     int buflen = 0;
     tlv *t;
-    off_t j = 0;
+    int j = 0;
     char *buf;
     assert(length != NULL);
     buflen++;			/* STORE_META_OK */
     buflen += sizeof(int);	/* size of header to follow */
     for (t = tlv_list; t; t = t->next)
 	buflen += sizeof(char) + sizeof(int) + t->length;
-    buflen++;			/* STORE_META_END */
     buf = xmalloc(buflen);
     buf[j++] = (char) STORE_META_OK;
     xmemcpy(&buf[j], &buflen, sizeof(int));
@@ -105,7 +112,6 @@ storeSwapMetaPack(tlv * tlv_list, int *length)
 	xmemcpy(&buf[j], t->value, t->length);
 	j += t->length;
     }
-    buf[j++] = (char) STORE_META_END;
     assert((int) j == buflen);
     *length = buflen;
     return buf;
@@ -119,7 +125,7 @@ storeSwapMetaUnpack(const char *buf, int *hdr_len)
     char type;
     int length;
     int buflen;
-    off_t j = 0;
+    int j = 0;
     assert(buf != NULL);
     assert(hdr_len != NULL);
     if (buf[j++] != (char) STORE_META_OK)
@@ -132,7 +138,7 @@ storeSwapMetaUnpack(const char *buf, int *hdr_len)
      */
     if (buflen <= (sizeof(char) + sizeof(int)))
 	    return NULL;
-    while (buflen - j > (sizeof(char) + sizeof(int))) {
+    while (buflen - j >= (sizeof(char) + sizeof(int))) {
 	type = buf[j++];
 	/* VOID is reserved, but allow some slack for new types.. */
 	if (type <= STORE_META_VOID || type > STORE_META_END + 10) {
